---
title: External dependencies
slug: '/guides/third-party-dependencies'
description: Learn how to define the contract between the dependency managers and Tuist.
---

:::warning Alpha
The integration of external dependencies is in alpha.
Be aware some APIs might change as we iterate the functionality with the feedback we get from users.
:::

# Dependencies.swift

Learn how to define the contract between the dependency managers (Cocoapods, Carthage and SPM) and Tuist.

## Integration status

- ✅ Carthage
- ✅ Swift Package Manager
- 🔴 Cocoapods

## How to get started

### Step 1: "Dependencies.swift"

1. Create `Dependencies.swift` in your project in `Tuist/` directory.
2. You can read more about `Dependencies.swift` [here](/manifests/dependencies/).
3. Run `tuist edit` to start editing your manifest files.
4. Add `import ProjectDescription` on the top of file.
5. Define your dependencies.

Example of project structure with added `Dependencies.swift` manifest file:

```bash
AwesomeProject
	|- App
	|- AppTests
	|- Tuist
		|- Dependencies.swift
	|- Project.swift
```

The snippet below shows an example `Dependencies.swift` manifest file:

```swift
import ProjectDescription

let dependencies = Dependencies(
    carthage: [
        .github(path: "Alamofire/Alamofire", requirement: .exact("5.0.4")),
    ],
    swiftPackageManager: [
        .remote(url: "https://github.com/Alamofire/Alamofire", requirement: .upToNextMajor(from: "5.0.0")),
    ],
    platforms: [.iOS]
)
```

### Step 2: "tuist dependencies" commands

1. Pull dependencies using `tuist dependencies fetch` command.
2. You can read more about commands [here](/commands/dependencies/).
3. Tuist will save dependencies in the `Tuist/Dependencies` directory.

The folder structure below shows how Tuist organizes dependencies:

```bash
Tuist
    |- Dependencies.swift # Manifest
    |- Dependencies
        |- graph.json # stores the serialized dependencies graph generated by `tuist dependencies`
        |- Lockfiles # stores the lockfiles generated by the dependencies resolution
            |- Carthage.resolved
            |- Package.resolved
            |- Podfile.lock # coming soon
        |- Carthage 
            |- Build # stores content of `Carthage/Build` directory generated by `Carthage`
                |- Alamofire.xcframework
                |- .Alamofire.version
        |- SwiftPackageManager
            |- .build # stores content of `.build/` directory generated by `Swift Package Manager`
                |- artifacts
                |- checkouts
                |- repositories
                |- manifest.db
                |- workspace-state.json
        |- Cocoapods # coming soon
            |- Pods # stores content of `Pods` directory generated by `CocoaPods`
                |- RxSwift
```

If you don't want to check in build artifacts you can update your `.gitignore`:

```bash
Tuist/Dependencies/graph.json # Avoid checking in the serialized dependencies graph generated by Tuist.
Tuist/Dependencies/Carthage # Avoid checking in build artifacts from Carthage dependencies.
Tuist/Dependencies/SwiftPackageManager # Avoid checking in build artifacts from Swift Package Manager dependencies.
Tuist/Dependencies/Cocoapods # Avoid checking in build artifacts from CocoaPods dependencies.
```

### Step 3: Link dependencies

Link pulled dependencies in your project manifest file.

1. Run `tuist edit` to start editing your manifest files.
2. Open `Project.swift`.
3. Define your project with dependencies.

The snippet below shows an example `Project.swift` manifest file:

```swift
import ProjectDescription

let project = Project(
    name: "App",
    organizationName: "tuist.io",
    targets: [
        Target(
            name: "App",
            platform: .iOS,
            product: .app,
            bundleId: "io.tuist.app",
            deploymentTarget: .iOS(targetVersion: "13.0", devices: .iphone),
            infoPlist: .default,
            sources: ["Targets/App/Sources/**"],
            dependencies: [
                .external(name: "Alamofire"),
            ]
        ),
    ]
)
```

## Merging Swift Packages into your project's graph

When Swift Packages are integrated into your project's graph, there are some heuristics in place to ensure the resulting graph is valid:
- Tuist defaults to using `iOS` as a platform when there's more than one platform defined in the `Dependencies.swift` and the package manifest file. This is currently a limitation of Tuist [because it does not support multi-platform targets](https://github.com/tuist/tuist/issues/397).
- Tuist defaults to using the deployment target defined in the `SwiftPackageManagerDependencies.deploymentTargets` property if the deployment target is not defined in the package manifest file. This is currently a limitation of Tuist [because it does not support automatic deployment targets](https://github.com/tuist/tuist/issues/3173).
- Tuist defaults to using the product type defined in the `SwiftPackageManagerDependencies.productTypes` property if the linking type is not defined in the package manifest file. If no product type is defined in `SwiftPackageManagerDependencies`, Tuist will default to a `.staticFramework`
